package com.sillycycle.bagleyd.hexagons;

/*
 * @(#)HexagonsCanvas.java
 *
 * Copyright 1994 - 2022  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Point;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.IOException;
import java.util.Random;
import javax.swing.JFrame;

/**
 * The <code>HexagonsCanvas</code> class does all the graphics for
 * <code>HexagonsFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

public class HexagonsCanvas extends Canvas {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");
	static final int MAX_ITERATIONS = 256;

	static final int TR = 0;
	static final int RIGHT = 1;
	static final int BR = 2;
	static final int BL = 3;
	static final int LEFT = 4;
	static final int TL = 5;
	static final int TOP = 6;
	static final int BOTTOM = 7;
	static final int COORD = 6;
	static final int TRBL = 0;
	static final int TLBR = 1;
	static final int ROW = 2;
	static final int ROW_TYPES = 3;
	static final int LOW = 0;
	static final int HIGH = 1;
	static final int SPACES = 2;
	static final int BLOCKED = -3;
	static final int SPACE = -2;
	static final int NORMAL = 1;
	static final int DOUBLE = 2;
	static final int INSTANT = 3;
	static final String DATAFILE = "hexagons.dat";
	static final String SYMBOL = ":";

	Color       background;
	Color       tileColor;
	int[]       tileOfPosition = null;
	int[]       startPosition = null;
	int[]       currentRow = new int[ROW_TYPES];
	int	 currentPosition = -1, currentPositionRow, currentSpace = -1;
	int	 currentDirection, lastDirection;
	int[]       spacePosition = new int[SPACES];
	int[]       startSpacePosition = new int[SPACES];
	int[]       spaceRow = new int[ROW_TYPES];
	int[]       startSpaceRow = new int[ROW_TYPES];
	boolean     started, cheat, corners;
	int	 base;
	int	 delay/*, numSlices*/;
	int	 sizeX, sizeY, sizeSize, sizeCenter;
	int	 pixX, pixY, pixNo;
	Point       coreSize;
	Point       offset;
	Point       tileSize, puzzleSize;
	Point       delta, puzzleOffset;

	JFrame      frame = null;
	int	 randomPaints = 0;
	boolean     firstPaint = false, resizePaint = false;
	boolean     movePaint = false, scanPaint = false, stackPaint = false;
	boolean     framePaint = false, allTilesPaint = false;
	boolean     selectPaint = false, releasePaint = false;
	boolean     mouseDown = false;
	boolean     focus = true;
	Random      generator;
	transient HexagonsStack undo = null, redo = null;
	String      stringSave = null;
	String[]    token = null;

	Point[][] hexagonUnit = new Point[HexagonsInterface.MAX_ORIENT][7];
	int[][] hexagonListX = new int[HexagonsInterface.MAX_ORIENT][7];
	int[][] hexagonListY = new int[HexagonsInterface.MAX_ORIENT][7];

	@Override
	protected String paramString() {
		return super.paramString() +
		",corners=" + corners +
		",sizeX=" + sizeX +
		",sizeY=" + sizeY +
		",base=" + base +
		",tileColor=" + tileColor +
		",background=" + background +
		",delay=" + delay;
	}

	void checkTiles() {
		if (sizeX < HexagonsInterface.MIN_TILES) {
			System.out.println(sizeX +
			" number of hexagons in a row incorrect, must be at least " +
			HexagonsInterface.MIN_TILES +
			", defaulting to " +
			HexagonsInterface.DEFAULT_TILESX);
			sizeX = HexagonsInterface.DEFAULT_TILESX;
		}
		if (sizeY < HexagonsInterface.MIN_TILES +
				((corners) ? 1 : 0)) {
			System.out.println(sizeY +
			" number of rows of hexagons incorrect, must be at least " +
			(HexagonsInterface.MIN_TILES + ((corners) ? 1 : 0)) +
			" in this case, defaulting to " +
			HexagonsInterface.DEFAULT_TILESY);
			sizeY = HexagonsInterface.DEFAULT_TILESY;
		}
		if (base < HexagonsInterface.MIN_BASE ||
				base > HexagonsInterface.MAX_BASE) {
			System.out.println(base +
			" base incorrect, must be between " +
			HexagonsInterface.MIN_BASE +
			" and " +
			HexagonsInterface.MAX_BASE +
			" inclusive, defaulting to " +
			HexagonsInterface.DEFAULT_BASE);
			base = HexagonsInterface.DEFAULT_BASE;
		}
		if (delay < 0) {
			System.out.println(delay +
			" delay cannot be negative, taking absolute value");
			delay = -delay;
		}
	}

	/* This is fast for small i, a -1 is returned for negative i. */
	static int sqrt(int i) {
		int j = 0;

		while (j * j <= i)
			j++;
		return (j - 1);
	}

	int positionFromRow(int rowPosition, int posRow) {
		return (posRow <= sizeY - 1) ?
			(sizeX * posRow + posRow * (posRow - 1) / 2 +
			rowPosition) : (sizeY * (2 * posRow - sizeY + 1) -
			sizeX * (posRow + 1) + 4 *
			sizeX * (sizeX - 1) / 2 - (2 * sizeX - posRow - 2) *
			(2 * sizeX - posRow - 1) / 2 + 1 + rowPosition);
	}

	int toPosition(int row, int trbl) {
		return ((row < sizeY) ?
			positionFromRow(trbl, row) :
			positionFromRow(trbl, row) - row + sizeY - 1);
	}

	int toRow(int pos) {
		return (pos <= sizeCenter) ?
			(1 + sqrt(1 + 8 * (pos + sizeX *
			(sizeX - 1) / 2))) / 2 - sizeX :
			sizeX + 2 * sizeY - 2 -
			(1 + sqrt(1 + 8 * (sizeSize - 1 +
			sizeX * (sizeX - 1) / 2 - pos))) / 2;
	}

	/* Passing row so there is no sqrt calculation again */
	int toTrBl(int pos, int posRow) {
		return (pos <= sizeCenter) ?
			(pos + sizeX * (sizeX - 1) / 2) -
			(posRow + sizeX) * (posRow + sizeX - 1) / 2 :
			sizeY + sizeX - 2 - (sizeSize - 1 +
			sizeX * (sizeX - 1) / 2 - pos -
			(2 * sizeY + sizeX - posRow - 2) *
			(2 * sizeY + sizeX - posRow - 3) / 2);
	}

	int toTlBr(int pos, int posRow) {
		return (pos <= sizeCenter) ?
			-(1 + pos + sizeX * (sizeX - 1) / 2) +
			(posRow + sizeX + 1) * (posRow + sizeX) / 2 :
			sizeY + sizeX - 1 + (sizeSize - 1 +
			sizeX * (sizeX - 1) / 2 - pos -
			(2 * sizeY + sizeX - posRow - 1) *
			(2 * sizeY + sizeX - posRow - 2) / 2);
	}

	int toDirHeading(int direction) {
		return (direction > 0 && direction < 4) ? HIGH : LOW;
	}

	int toRowType(int direction) {
		switch (direction) {
		case TR:
		case BL:
			return TRBL;
		case TL:
		case BR:
			return TLBR;
		case LEFT:
		case RIGHT:
			return ROW;
		default:
			return -1;
		}
	}

	int cartesianX(int pos, int row) {
		return tileSize.x / 4 - 1 + (2 * toTrBl(pos, row) +
			sizeY - row) * offset.x / 2 - 2 +
			puzzleOffset.x;
	}

	int cartesianY(int row) {
		return row * (3 * tileSize.y / 2 + delta.y) +
			delta.y + puzzleOffset.y;
	}

	int tileNFromSpace(int n, int rowType, int direction) {
		int pos = spacePosition[HIGH];

		if (rowType == ROW) /* This one is easy */
			return (pos + ((direction == HIGH) ? -n : n));
		int row = toRow(pos);
		int trbl = toTrBl(pos, row);
		int spaceOffset;

		if (direction == HIGH) {
			spaceOffset = -n;
		} else {
			spaceOffset = n;
		}
		switch (rowType) {
		case TRBL:
			return toPosition(row + spaceOffset, trbl);
		case TLBR:
			return toPosition(row + spaceOffset,
				trbl + spaceOffset);
		default: /* ROW */
			return (pos + spaceOffset);
		}
	}

	int tileNextToSpace(int rowType, int direction) {
		if (direction == HIGH) {
			switch (rowType) {
			case TRBL:
				return ((spaceRow[ROW] < sizeY) ?
					spacePosition[HIGH] - sizeX -
					spaceRow[ROW] + 1 :
					spacePosition[HIGH] - 2 * sizeY -
					sizeX + spaceRow[ROW] + 2);
			case TLBR:
				return ((spaceRow[ROW] < sizeY) ?
					spacePosition[HIGH] - sizeX -
					spaceRow[ROW] :
					spacePosition[HIGH] - 2 * sizeY -
					sizeX + spaceRow[ROW] + 1);
			default: /* ROW */
				return (spacePosition[HIGH] - 1);
			}
		}
		/* direction == LOW */
		switch (rowType) {
		case TRBL:
			return ((spaceRow[ROW] < sizeY - 1) ?
				spacePosition[HIGH] + sizeX +
				spaceRow[ROW] :
				spacePosition[HIGH] + 2 * sizeY +
				sizeX - spaceRow[ROW] - 3);
		case TLBR:
			return ((spaceRow[ROW] < sizeY - 1) ?
				spacePosition[HIGH] + sizeX +
				spaceRow[ROW] + 1 :
				spacePosition[HIGH] + 2 * sizeY +
				sizeX - spaceRow[ROW] - 2);
		default: /* ROW */
			return (spacePosition[HIGH] + 1);
		}
	}

	public boolean checkSolved() {
		int i;

		for (i = 1; i < sizeSize - ((corners) ?
				HexagonsInterface.CORNERS :
				HexagonsInterface.NOCORN); i++)
			if (tileOfPosition[i - 1] != i)
				return false;
		started = false;
		return true;
	}

	boolean bresenhamLine(int x0, int y0, int x1, int y1) {
		int pix = 0;
		int ex = x1 - x0;
		int ey = y1 - y0;
		int dx, dy, error;

		pixNo++;
		if (ex > 0) {
			dx = 1;
		} else if (ex < 0) {
			dx = -1;
			ex = -ex;
		} else {
			dx = 0;
		}
		if (ey > 0) {
			dy = 1;
		} else if (ey < 0) {
			dy = -1;
			ey = -ey;
		} else {
			dy = 0;
		}
		pixX = x0;
		pixY = y0;
		if (ex > ey) {
			error = 2 * ey - ex;
			while (pix != pixNo) {
				if (error >= 0) {
					error -= 2 * ex;
					pixY += dy;
				}
				error += 2 * ey;
				pixX += dx;
				pix++;
				if (pixX == x1)
					return true;
			}
			return false;
		}
		error = 2 * ex - ey;
		while (pix != pixNo) {
			if (error >= 0) {
				error -= 2 * ey;
				pixX += dx;
			}
			error += 2 * ex;
			pixY += dy;
			pix++;
			if (pixY == y1)
				return true;
		}
		return false;
	}

	static void fill3DPolygon(Graphics g, int startX, int startY,
			int[] listX, int[] listY, int n, boolean raised,
			boolean corners) {
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		Color c = g.getColor();
		Color brighter = c.brighter();
		Color darker = c.darker();

		if (!raised) {
			g.setColor(darker);
		}
		g.fillPolygon(tempListX, tempListY, n);
		if (!corners) {
			g.drawLine(tempListX[4], tempListY[4],
				tempListX[5], tempListY[5]);
			g.drawLine(tempListX[1], tempListY[1],
				tempListX[2], tempListY[2]);
		}
		if (!raised && corners) {
			g.setColor(darker);
			g.drawLine(tempListX[4], tempListY[4],
				tempListX[5], tempListY[5]);
			g.drawLine(tempListX[4] + 1, tempListY[4],
				tempListX[5] + 1, tempListY[5]);
		}
		g.setColor(raised ? brighter : darker);
		g.drawLine(tempListX[5], tempListY[5],
			tempListX[0], tempListY[0]);
		g.drawLine(tempListX[0], tempListY[0],
			tempListX[1], tempListY[1]);
		g.drawLine(tempListX[0], tempListY[0] + 1,
			tempListX[1], tempListY[1] + 1);
		if (corners) {
			g.drawLine(tempListX[5], tempListY[5] + 1,
				tempListX[0], tempListY[0] + 1);
			g.drawLine(tempListX[5], tempListY[5] + 2,
				tempListX[0], tempListY[0] + 2);
		} else {
			g.drawLine(tempListX[5] + 1, tempListY[5],
				tempListX[0] + 1, tempListY[0]);
			g.drawLine(tempListX[5] + 2, tempListY[5],
				tempListX[0] + 2, tempListY[0]);
		}
		g.setColor(raised ? darker : c);
		g.drawLine(tempListX[2], tempListY[2],
			tempListX[3], tempListY[3]);
		g.drawLine(tempListX[3], tempListY[3],
			tempListX[4], tempListY[4]);
		g.drawLine(tempListX[3], tempListY[3] - 1,
			tempListX[4], tempListY[4] - 1);
		if (corners) {
			g.drawLine(tempListX[1], tempListY[1],
				tempListX[2], tempListY[2]);
			g.drawLine(tempListX[1] - 1, tempListY[1],
				tempListX[2] - 1, tempListY[2]);
			g.drawLine(tempListX[2], tempListY[2] - 1,
				tempListX[3], tempListY[3] - 1);
			g.drawLine(tempListX[2], tempListY[2] - 2,
				tempListX[3], tempListY[3] - 2);
			if (raised) {
				g.setColor(brighter);
				g.drawLine(tempListX[4], tempListY[4],
					tempListX[5], tempListY[5]);
				g.drawLine(tempListX[4] + 1, tempListY[4],
					tempListX[5] + 1, tempListY[5]);
			}
		} else {
			g.drawLine(tempListX[2] - 1, tempListY[2],
				tempListX[3] - 1, tempListY[3]);
			g.drawLine(tempListX[2] - 2, tempListY[2],
				tempListX[3] - 2, tempListY[3]);
		}
		g.setColor(c);
	}

	static void drawShadow(Graphics g, int startX, int startY,
			int[] listX, int[] listY, int n, boolean corners) {
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		g.drawLine(tempListX[4], tempListY[4],
			tempListX[5], tempListY[5]);
		g.drawLine(tempListX[5], tempListY[5],
			tempListX[0], tempListY[0]);
		g.drawLine(tempListX[0], tempListY[0],
			tempListX[1], tempListY[1]);
		if (corners)
			g.drawLine(tempListX[5], tempListY[5] + 1,
				tempListX[0], tempListY[0] + 1);
		else
			g.drawLine(tempListX[5] + 1, tempListY[5],
				tempListX[0] + 1, tempListY[0]);
	}

	static void fillPolygon(Graphics g, int startX, int startY,
			int[] listX, int[] listY, int n) {
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		g.fillPolygon(tempListX, tempListY, n);
		g.drawPolygon(tempListX, tempListY, n);
	}

	void drawTile(int pos, boolean blank, boolean erase,
			int pressedOffset, int offsetX, int offsetY) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			int dx, dy, row = toRow(pos);
			int startOffsetX = 0, startOffsetY = 0;
			int orient = (corners) ? HexagonsInterface.CORNERS :
				HexagonsInterface.NOCORN;

			dx = cartesianX(pos, row) + offsetX + pressedOffset;
			dy = cartesianY(row) + offsetY + pressedOffset;

			Color fillColor, textColor;

			if (erase) {
				fillColor = background;
				textColor = background;
			} else {
				fillColor = tileColor;
				textColor = background.darker();
			}
			if (!corners) {
				startOffsetX = -offset.x / 4;
				startOffsetY = (tileSize.y + delta.y) / 4;
			}

			if (blank) {
				g.setColor(fillColor);
				fillPolygon(g,
					dx + startOffsetX, dy + startOffsetY,
					hexagonListX[orient],
					hexagonListY[orient], 6);
			} else {
				if (pressedOffset != 0) {
					g.setColor(fillColor.darker());
					drawShadow(g,
						dx + startOffsetX - pressedOffset,
						dy + startOffsetY - pressedOffset,
						hexagonListX[orient],
						hexagonListY[orient], 6,
						corners);
				}
				g.setColor(fillColor);
				fill3DPolygon(g,
					dx + startOffsetX, dy + startOffsetY,
					hexagonListX[orient],
					hexagonListY[orient], 6,
					pressedOffset == 0, corners);
			}
			if (!blank) {
				int i, tile = tileOfPosition[pos];
				int digitOffsetX = 0, digitOffsetY, fontSize;

				g.setColor(textColor);
				fontSize = java.lang.Math.min(tileSize.x,
					tileSize.y) / 2;
				g.setFont(new Font("arial", Font.PLAIN,
					fontSize));
				i = tile;
				while (i >= 1) {
					i /= base;
					digitOffsetX += (fontSize / 3);
				}
				digitOffsetY = fontSize / 2 + 1 -
					tileSize.y / 8;
				g.drawString(Integer.toString(tile, base).toUpperCase(),
					dx - digitOffsetX,
					dy + tileSize.y + digitOffsetY);
			}
		} finally {
			g.dispose();
		}
	}

	void drawAllTiles() {
		int k;

		for (k = 0; k < sizeSize; k++) {
			drawTile(k, (tileOfPosition[k] <= 0),
				(tileOfPosition[k] <= 0), 0, 0, 0);
		}
	}

	int findSpaceType(int p1, int p2, int r1, int r2) {
		int pos1 = p1, pos2 = p2, row1 = r1, row2 = r2;

		if (row1 == row2 && (pos1 == pos2 + 1 || pos1 == pos2 - 1))
			return (ROW);
		else if (row1 == row2 - 1) {
			int tmp;

			tmp = row2; /* Swap rows */
			row2 = row1;
			row1 = tmp;
			tmp = pos2; /* Swap pos */
			pos2 = pos1;
			pos1 = tmp;
		}
		if (row1 == row2 + 1) {
			if (row1 <= sizeY - 1) {
				if (pos2 == pos1 - sizeX - row1)
					return (TLBR);
				else if (pos2 == pos1 - sizeX - row1 + 1)
					return (TRBL);
			} else {	/* row1 > size - 1 */
				if (pos2 == pos1 - 2 * sizeY - sizeX + row1 + 1)
					return (TLBR);
				else if (pos2 == pos1 - 2 * sizeY - sizeX + row1 + 2)
					return (TRBL);
			}
		}
		return (-1);
	}

	void findMovableTile(int pos, int posRow, int spaceType, int side) {
		switch (spaceType) {
		case TRBL:
			if (side == HIGH) {
				currentPositionRow = posRow;
				currentPosition = pos + 1;
			} else {	/* side == LOW */
				currentPositionRow = posRow - 1;
				currentPosition = (posRow <= sizeY - 1) ?
					pos - sizeX - posRow :
					pos - 2 * sizeY - sizeX +
					posRow + 1;
			}
			break;
		case TLBR:
			if (side == HIGH) {
				currentPositionRow = posRow;
				currentPosition = pos - 1;
			} else {	/* side == LOW */
				currentPositionRow = posRow - 1;
				currentPosition = (posRow <= sizeY - 1) ?
					pos - sizeX - posRow + 1 :
					pos - 2 * sizeY - sizeX +
					posRow + 2;
			}
			break;
		case ROW:
			if (side == HIGH) {
				currentPositionRow = posRow + 1;
				currentPosition = (posRow < sizeY - 1) ?
					pos + sizeX + posRow :
					pos + 2 * sizeY + sizeX -
					posRow - 3;
			} else {	/* side == LOW */
				currentPositionRow = posRow - 1;
				currentPosition = (posRow <= sizeY - 1) ?
					pos - sizeX - posRow :
					pos - 2 * sizeY - sizeX +
					posRow + 1;
			}
			break;
		default:
			System.out.println("findMovableTile: spaceType " +
				spaceType);
		}
		if (currentPosition < 0)
			System.out.println("currentPosition " + currentPosition +

				", side " + side +
				", pos " + pos +
				", posRow " + posRow +
				", spaceType " + spaceType);
	}

	int nextToWall(int pos, int posRow, int spaceType) {
		switch (spaceType) {
		case TRBL:
			if (posRow < sizeY && pos == sizeX * posRow +
					posRow * (posRow - 1) / 2)
				return (HIGH);
			else if (posRow >= sizeY && pos == sizeY *
				(2 * posRow - sizeY + 3) - sizeX *
				posRow - 2 - posRow +
				4 * sizeX * (sizeX - 1) / 2 -
				(2 * sizeX - posRow - 2) *
				(2 * sizeX - posRow - 1) / 2)
				return (LOW);
			else
				return (-1);
		case TLBR:
			if (posRow < sizeY && pos == sizeX * (posRow +
					1) + posRow * (posRow + 1) / 2 - 1)
				return (HIGH);
			else if (posRow >= sizeY && pos == sizeY *
				(2 * posRow - sizeY + 1) - sizeX *
				(posRow + 1) + 1 + 4 * sizeX *
				(sizeX - 1) / 2 - (2 * sizeX - posRow - 2) *
				(2 * sizeX - posRow - 1) / 2)
				return (LOW);
			else
				return (-1);
		case ROW:
			if (posRow == 0)
				return (HIGH);
			else if (posRow == 2 * (sizeY - 1))
				return (LOW);
			else
				return (-1);
		default:
		}
		return (-2);		/*Unknown space formation. */
	}

	int movableCornerTile() {
		/* (Clicked on a space?) */
		if (currentPosition == spacePosition[LOW] ||
				currentPosition == spacePosition[HIGH]) {
			return SPACE;
		}
		if (findTileTriangle(currentPosition, spacePosition[HIGH],
				spacePosition[LOW], currentRow[ROW],
				spaceRow[HIGH], spaceRow[LOW])) {
			return 0;
		}
		return BLOCKED;
	}

	int movableNoCornTile() {
		int rowType = BLOCKED, l;

		/* Are the spaces in a "row" with the mouse click?
		 * (If two, then one clicked on a space). */
		for (l = 0; l < ROW_TYPES; l++) {
			if (currentRow[l] == spaceRow[l]) {
				if (rowType == BLOCKED) {
					rowType = l;
				} else {
					return SPACE;
				}
			}
		}
		return rowType;
	}

	int countTiles(int rowType) {
		switch (rowType) {
		case TLBR:
			return java.lang.Math.abs(spaceRow[ROW] -
				currentRow[ROW]);
		case TRBL:
			return java.lang.Math.abs(spaceRow[ROW] -
				currentRow[ROW]);
		case ROW:
			return java.lang.Math.abs(spaceRow[TRBL] -
				currentRow[TRBL]);
		default:
			System.out.println("countTiles: rowType " + rowType);
		}
		return 0;
	}

	static int dirTiles(int dir, int rowType) {
		switch (rowType) {
		case TLBR:
			return (dir == HIGH) ? BR : TL;
		case TRBL:
			return (dir == HIGH) ? BL : TR;
		case ROW:
			return (dir == HIGH) ? RIGHT : LEFT;
		default:
			System.out.println("dirTiles: rowType " + rowType);
		}
		return dir;
	}

	void animateSlide(int dir, int fast, boolean logMoves) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			int aTile, numTiles, direction = dir, rowType;
			int fillBeginPos;
			int pos, posNext;

			rowType = movableNoCornTile();
			if (rowType < 0) {
				System.out.println(
					"animateSlide: rowType " + rowType);
				numTiles = 0;
			} else {
				numTiles = countTiles(rowType);
				direction = dirTiles(dir, rowType);
			}
			if (numTiles < 0)
				numTiles = -numTiles;
			fillBeginPos = tileNFromSpace(numTiles, rowType, dir);
			{
				int fromdx, fromdy;
				int todx, tody;
				int dx, dy, row;
				int[] rowPos = new int[numTiles + 1];
				int i, k;
				boolean slideDone = false;

				pixNo = 0;
				pixX = 0;
				pixY = 0;
				rowPos[1] = tileNFromSpace(1, rowType, dir);
				rowPos[0] = spacePosition[HIGH];
				for (i = 1; i < numTiles; i++) {
					rowPos[i + 1] = tileNFromSpace(i + 1,
						rowType, dir);
				}
				row = toRow(rowPos[1]);
				fromdx = cartesianX(rowPos[1], row);
				fromdy = cartesianY(row);
				row = toRow(rowPos[0]);
				todx = cartesianX(rowPos[0], row);
				tody = cartesianY(row);
				dx = todx - fromdx;
				dy = tody - fromdy;
				i = 0;
				while (!slideDone) {
					for (k = 1; k < numTiles + 1; k++)
						drawTile(rowPos[k],
							true, true, 0,
							pixX, pixY);
					slideDone = bresenhamLine(0, 0, dx, dy);
					for (k = 1; k < numTiles + 1; k++)
						drawTile(rowPos[k],
							false, false, 0,
							pixX, pixY);
					if (i % 8 == 0) {
						try {
							Thread.sleep(delay / fast);
						} catch (InterruptedException e) {
							// ignore
						}
					}
					i++;
				}
			}
			for (aTile = 0; aTile < numTiles; aTile++) {
				if (logMoves) {
					((HexagonsFrame) frame).callback(
						HexagonsInterface.ACTION_MOVED);
					undo.setMove(direction);
					redo.flushMoves();
				}
			}
			pos = spacePosition[HIGH];
			for (aTile = 0; aTile < numTiles; aTile++) {
				posNext = tileNFromSpace(aTile + 1, rowType, dir);
				tileOfPosition[pos] =
					tileOfPosition[posNext];
				pos = posNext;
			}
			tileOfPosition[fillBeginPos] =
				spacePosition[HIGH];
			spacePosition[HIGH] = fillBeginPos;
			tileOfPosition[fillBeginPos] = 0;
			spaceRow[ROW] = toRow(fillBeginPos);
			spaceRow[TRBL] =
				toTrBl(fillBeginPos, spaceRow[ROW]);
			spaceRow[TLBR] =
				toTlBr(fillBeginPos, spaceRow[ROW]);
		} finally {
			g.dispose();
		}
	}

	void resetTiles() {
		int i;

		/* sizeSize = 3 * size * (size - 1) + 1; */
		sizeSize = sizeY * sizeY + (sizeX - 1) * (sizeY * 2 - 1);
		sizeCenter = (sizeSize - 1) / 2;
		tileOfPosition = new int[sizeSize];
		startPosition = new int[sizeSize];
		/*spacePosition[HIGH] = sizeSize - 1;
		if (corners) {
			spaceRow[HIGH] = 2 * sizeY - 2;
			if (sizeX > 1 || sizeY > 1) {
				spacePosition[LOW] = sizeSize - 2;
				spaceRow[LOW] = 2 * sizeY -
					((sizeX == 1) ? 3 : 2);
				tileOfPosition[sizeSize - 2] = -1;
			}
		} else {
			spaceRow[ROW] = spaceRow[TRBL] = 2 * sizeY - 2;
			spaceRow[TLBR] = sizeY - 1;
		}
		tileOfPosition[sizeSize - 1] = 0;
		for (i = 0; i < sizeSize - 1 - ((corners) ? 1 : 0); i++)
			tileOfPosition[i] = i + 1;*/
		if (corners) {
			if (sizeX > 1 || sizeY > 1) {
				spacePosition[HIGH] = 1;
				spaceRow[HIGH] = (sizeX == 1) ? 1 : 0;
				spacePosition[LOW] = 0;
				spaceRow[LOW] = 0;
				tileOfPosition[0] = -1;
				tileOfPosition[1] = 0;
			} else {
				spacePosition[HIGH] = 0;
				spaceRow[HIGH] = 0;
				tileOfPosition[0] = 0;
			}
		} else {
			spacePosition[HIGH] = 0;
			spaceRow[ROW] = spaceRow[TRBL] = 0;
			spaceRow[TLBR] = sizeX - 1;
			tileOfPosition[0] = 0;
		}
		if (corners && sizeX == 1 && sizeY == 3) {
			tileOfPosition[2] = 1;
			tileOfPosition[3] = 5;
			tileOfPosition[4] = 4;
			tileOfPosition[5] = 2;
			tileOfPosition[6] = 6;
			tileOfPosition[7] = 3;
			tileOfPosition[8] = 7;
		} else if (corners && sizeX == 2 && sizeY == 2) {
			tileOfPosition[2] = 4;
			tileOfPosition[3] = 3;
			tileOfPosition[4] = 1;
			tileOfPosition[5] = 5;
			tileOfPosition[6] = 2;
		} else {
			for (i = 0; i < sizeSize - 1 - ((corners) ? 1 : 0);
					i++)
				tileOfPosition[i + 1 + ((corners) ? 1 : 0)] =
					i + 1;
		}
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();
		currentPosition = -1;
		currentDirection = -1;
		started = false;
		cheat = false;
		randomPaints = 0;
	}

	void drawFrame(Graphics g, boolean show) {
		int sumX, sumY, sumX4, sum3X4, sumY2, offsetX, offsetY;
		int[] tempListX = new int[6];
		int[] tempListY = new int[6];

		if (g == null)
			return;
		sumX = (sizeX + sizeY) * offset.x -
			tileSize.x / 2 - 2 * delta.x - 1;
		sumY = sizeY * (3 * tileSize.y + 2 *
			delta.y) - tileSize.y - 1;
		offsetX = puzzleOffset.x;
		offsetY = puzzleOffset.y;
		sumX4 = sizeY * offset.x / 2 + offsetX -
			tileSize.x / 4 + 1;
		sum3X4 = sizeX * offset.x + sumX4 - 12;
		sumY2 = sumY / 2 + offsetY;
		sumX += offsetX + sizeX / 2;
		sumY += offsetY;
		offsetX += -1 - sizeX / 2;
		g.setColor(background.darker().darker());
		if (offsetY > 0)
			g.fillRect(0, 0, coreSize.x, offsetY);
		if (sumY < coreSize.y)
			g.fillRect(0, sumY + 1, coreSize.x,
				coreSize.y - sumY);
		if (offsetX > 0)
			g.fillRect(0, 0, offsetX, coreSize.y);
		if (sumX < coreSize.x)
			g.fillRect(sumX, 0, coreSize.x - sumX,
				coreSize.y);
		tempListX[0] = offsetX;
		tempListY[0] = offsetY;
		tempListX[1] = sumX4;
		tempListY[1] = offsetY;
		tempListX[2] = offsetX;
		tempListY[2] = sumY2;
		g.fillPolygon(tempListX, tempListY, 3);
		tempListX[0] = sumX;
		tempListY[0] = offsetY;
		tempListX[1] = sumX;
		tempListY[1] = sumY2;
		tempListX[2] = sum3X4;
		tempListY[2] = offsetY;
		g.fillPolygon(tempListX, tempListY, 3);
		tempListX[0] = offsetX;
		tempListY[0] = sumY;
		tempListX[1] = offsetX;
		tempListY[1] = sumY2;
		tempListX[2] = sumX4;
		tempListY[2] = sumY;
		g.fillPolygon(tempListX, tempListY, 3);
		tempListX[0] = sumX;
		tempListY[0] = sumY;
		tempListX[1] = sum3X4;
		tempListY[1] = sumY;
		tempListX[2] = sumX;
		tempListY[2] = sumY2;
		g.fillPolygon(tempListX, tempListY, 3);
		g.drawLine(offsetX, sumY, sumX4, sumY);
		g.drawLine(sumX, sumY, sum3X4, sumY);
		if (show) { /* 3DFrame */
			Color brighterColor = (focus) ?
				background.brighter() : background;
			Color darkerColor = (focus) ?
				background.darker() : background;

			tempListX[0] = sum3X4;
			tempListY[0] = offsetY;
			tempListX[1] = sumX;
			tempListY[1] = sumY2;
			tempListX[2] = sum3X4;
			tempListY[2] = sumY;
			tempListX[3] = sumX4;
			tempListY[3] = sumY;
			g.setColor(brighterColor);
			g.drawPolyline(tempListX, tempListY, 4);
			tempListX[0] = sumX4;
			tempListY[0] = sumY;
			tempListX[1] = offsetX;
			tempListY[1] = sumY2;
			tempListX[2] = sumX4;
			tempListY[2] = offsetY;
			tempListX[3] = sum3X4;
			tempListY[3] = offsetY;
			g.setColor(darkerColor);
			g.drawPolyline(tempListX, tempListY, 4);
			tempListX[0] = sum3X4 - 1;
			tempListY[0] = offsetY;
			tempListX[1] = sumX - 1;
			tempListY[1] = sumY2;
			tempListX[2] = sum3X4 - 1;
			tempListY[2] = sumY;
			tempListX[3] = sumX4 + 1;
			tempListY[3] = sumY;
			g.setColor(brighterColor);
			g.drawPolyline(tempListX, tempListY, 4);
			tempListX[0] = sumX4 + 1;
			tempListY[0] = sumY;
			tempListX[1] = offsetX + 1;
			tempListY[1] = sumY2;
			tempListX[2] = sumX4 + 1;
			tempListY[2] = offsetY;
			tempListX[3] = sum3X4 - 1;
			tempListY[3] = offsetY;
			g.setColor(darkerColor);
			g.drawPolyline(tempListX, tempListY, 4);
			tempListX[0] = sum3X4;
			tempListY[0] = offsetY - 1;
			tempListX[1] = sumX + 1;
			tempListY[1] = sumY2;
			tempListX[2] = sum3X4;
			tempListY[2] = sumY + 1;
			tempListX[3] = sumX4;
			tempListY[3] = sumY + 1;
			g.setColor(brighterColor);
			g.drawPolyline(tempListX, tempListY, 4);
			tempListX[0] = sumX4;
			tempListY[0] = sumY + 1;
			tempListX[1] = offsetX - 1;
			tempListY[1] = sumY2;
			tempListX[2] = sumX4;
			tempListY[2] = offsetY - 1;
			tempListX[3] = sum3X4;
			tempListY[3] = offsetY - 1;
			g.setColor(darkerColor);
			g.drawPolyline(tempListX, tempListY, 4);
		} else {
			tempListX[0] = sumX4;
			tempListY[0] = offsetY;
			tempListX[1] = sum3X4;
			tempListY[1] = offsetY;
			tempListX[2] = sumX;
			tempListY[2] = sumY2;
			tempListX[3] = sum3X4;
			tempListY[3] = sumY;
			tempListX[4] = sumX4;
			tempListY[4] = sumY;
			tempListX[5] = offsetX;
			tempListY[5] = sumY2;
			g.setColor(background);
			g.fillPolygon(tempListX, tempListY, 6);
		}
	}

	public void drawFrame(boolean show) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			drawFrame(g, show);
		} finally {
			g.dispose();
		}
	}

	void moveNoTiles() {
		((HexagonsFrame) frame).callback(
			HexagonsInterface.ACTION_IGNORE);
	}

	void moveCornerTiles(int from, int posRow, int space, int fast) {
		int fromTile, nextPos, row;
		int dx, dy, s0x, s0y, s1x, s1y, cx, cy;

		fromTile = tileOfPosition[from];
		nextPos = spacePosition[space];
		row = toRow(from);
		dx = cartesianX(from, row);
		dy = cartesianY(row);
		row = toRow(spacePosition[space]);
		s0x = cartesianX(spacePosition[space], row);
		s0y = cartesianY(row);
		row = toRow(spacePosition[(space == HIGH) ? LOW : HIGH]);
		s1x = cartesianX(spacePosition[(space == HIGH) ? LOW : HIGH],
			row);
		s1y = cartesianY(row);
		cx = (s0x + s1x - 2 * dx) / 3;
		cy = (s0y + s1y - 2 * dy) / 3;
		if (fast == INSTANT) {
			drawTile(from, true, true, 0, 0, 0);
			drawTile(from, false, false, 0, cx, cy);
			if (((HexagonsFrame) frame).getToggleSound()) {
				((HexagonsFrame) frame).playBumpAudio();
			}
			drawTile(from, true, true, 0, cx, cy);
			drawTile(from, false, false, 0, s0x - dx, s0y - dy);
		} else {
			boolean slideDone = false;
			int i = 0;

			pixNo = 0;
			pixX = 0;
			pixY = 0;
			while (!slideDone) {
				drawTile(from, true, true, 0, pixX, pixY);
				slideDone = bresenhamLine(0, 0, cx, cy);
				drawTile(from, false, false, 0, pixX, pixY);
				if (i % 8 == 0) {
					try {
						Thread.sleep(delay / fast);
					} catch (InterruptedException e) {
						// ignore
					}
				}
				i++;
			}
			if (((HexagonsFrame) frame).getToggleSound()) {
				((HexagonsFrame) frame).playBumpAudio();
			}
			slideDone = false;
			i = 0;
			pixNo = 0;
			while (!slideDone) {
				drawTile(from, true, true, 0, pixX, pixY);
				slideDone = bresenhamLine(cx, cy, s0x - dx, s0y - dy);
				drawTile(from, false, false, 0, pixX, pixY);
				if (i % 8 == 0) {
					try {
						Thread.sleep(delay / fast);
					} catch (InterruptedException e) {
						// ignore
					}
				}
				i++;
			}
		}
		tileOfPosition[from] =
			tileOfPosition[spacePosition[space]];
		spacePosition[space] = from;
		spaceRow[space] = posRow;
		tileOfPosition[nextPos] = fromTile;
		if (((HexagonsFrame) frame).getToggleSound()) {
			((HexagonsFrame) frame).playBumpAudio();
		}
		if (fast != INSTANT) {
			try {
				Thread.sleep(10 * delay / fast);
			} catch (InterruptedException e) {
				// ignore
			}
		}
	}

	void moveNoCornTiles(int from, int posRow, int fast) {
		if (fast != INSTANT && delay > 0) {
			currentPosition = from;
			currentRow[ROW] = toRow(from);
			currentRow[TRBL] = toTrBl(from, currentRow[ROW]);
			currentRow[TLBR] = toTlBr(from, currentRow[ROW]);
			if (currentPosition < spacePosition[HIGH]) {
				animateSlide(HIGH, fast, false);
			} else if (currentPosition > spacePosition[HIGH]) {
				animateSlide(LOW, fast, false);
			} else
				return;
			currentPosition = -1;
		} else {
			int tempTile;

			tempTile = tileOfPosition[from];
			tileOfPosition[from] =
				tileOfPosition[spacePosition[HIGH]];
			tileOfPosition[spacePosition[HIGH]] = tempTile;
			drawTile(spacePosition[HIGH],
				false, false, 0, 0, 0);
			spacePosition[HIGH] = from;
			spaceRow[ROW] = posRow;
			spaceRow[TRBL] = toTrBl(from, posRow);
			spaceRow[TLBR] = toTlBr(from, posRow);
			drawTile(spacePosition[HIGH],
				true, true, 0, 0, 0);
		}
		if (((HexagonsFrame) frame).getToggleSound())
			((HexagonsFrame) frame).playBumpAudio();
	}

	boolean findTileTriangle(int pI, int pJ, int pK,
			int rI, int rJ, int rK) {
		int temp = 0, k = 0, row1 = 0, row2 = 0, pos;
		boolean found = true;

		if (rI == rJ) {
			if (pI == pJ - 1)
				temp = pJ;
			else if (pI == pJ + 1)
				temp = pI;
			else
				found = false;
			k = pK;
			row1 = rI;
			row2 = rK;
		} else if (rJ == rK) {
			if (pJ == pK - 1)
				temp = pK;
			else if (pJ == pK + 1)
				temp = pJ;
			else
				found = false;
			k = pI;
			row1 = rJ;
			row2 = rI;
		} else if (rK == rI) {
			if (pK == pI - 1)
				temp = pI;
			else if (pK == pI + 1)
				temp = pK;
			else
				found = false;
			k = pJ;
			row1 = rK;
			row2 = rJ;
		}
		if (!found)
			return false;
		pos = -1;
		if (row1 == row2 + 1) {
			if (row1 <= sizeY - 1)
				pos = temp - sizeX - row1;
			else		/* row1 > sizeY - 1 */
				pos = temp - 2 * sizeY - sizeX + row1 + 1;
		} else if (row1 == row2 - 1) {
			if (row1 < sizeY - 1)
				pos = temp + sizeX + row1;
			else		/* row1 >= sizeY - 1 */
				pos = temp + 2 * (sizeY - 1) + (sizeX - 1) - row1;
		}
		return (k == pos);
	}

	int findDir(int posTile, int posSpace, int rowTile, int rowSpace) {
		if (rowTile == rowSpace) {
			if (posTile > posSpace)
				return LEFT;
			return RIGHT;
		} else if (toTrBl(posTile, rowTile) == toTrBl(posSpace, rowSpace)) {
			if (posTile > posSpace)
				return TR;
			return BL;
		} else {
			/* if (toTlBr(posTile, rowTile) == toTlBr(posSpace, rowSpace)) */
			if (posTile > posSpace)
				return TL;
			return BR;
		}
	}

	boolean withinFrame(int x, int y, int dx, int dy) {
		return (x < dx + tileSize.x / 2 &&
			x > dx - tileSize.x / 2 &&
			tileSize.y * (x - dx) < tileSize.x * (y - dy) &&
			tileSize.y * (dx - x) < tileSize.x * (y - dy) &&
			tileSize.y * (x - dx + 2 * tileSize.x) >
				tileSize.x * (y - dy) &&
			tileSize.y * (dx - x + 2 * tileSize.x) >
				tileSize.x * (y - dy));
	}

	boolean checkMoveCornerTilesDir(int direction) {
		int spaceType, highest, side = -1;

		highest = (spacePosition[HIGH] >
			spacePosition[LOW]) ? HIGH : LOW;
		spaceType = findSpaceType(spacePosition[HIGH], spacePosition[LOW],
			spaceRow[HIGH], spaceRow[LOW]);
		switch (spaceType) {
		case TRBL:
			if (direction == TR || direction == BL)
				return false;
			side = nextToWall(spacePosition[highest],
				spaceRow[highest], spaceType);
			if (side == -1) {
				side = (direction == TL || direction == LEFT) ?
					HIGH : LOW;
			} else {
				if ((side == HIGH && direction == RIGHT) ||
				(side == HIGH && direction == BR) ||
				(side == LOW && direction == LEFT) ||
				(side == LOW && direction == TL))
					return false;
			}
			currentSpace = (direction == BR || direction == LEFT) ?
				HIGH : LOW;
			break;
		case TLBR:
			if (direction == TL || direction == BR)
				return false;
			side = nextToWall(spacePosition[highest],
				spaceRow[highest], spaceType);
			if (side == -1) {
				side = (direction == TR || direction == RIGHT) ?
					HIGH : LOW;
			} else {
				if ((side == LOW && direction == TR) ||
				(side == LOW && direction == RIGHT) ||
				(side == HIGH && direction == BL) ||
				(side == HIGH && direction == LEFT))
					return false;
			}
			currentSpace = (direction == BL || direction == RIGHT) ?
					HIGH : LOW;
			break;
		case ROW:
			if (direction == LEFT || direction == RIGHT)
				return false;
			side = nextToWall(spacePosition[highest],
				spaceRow[highest], spaceType);
			if (side == -1) {
				side = (direction == TL || direction == TR) ?
					HIGH : LOW;
			} else {
				if ((side == LOW && direction == TR) ||
				(side == HIGH && direction == BR) ||
				(side == HIGH && direction == BL) ||
				(side == LOW && direction == TL))
					return false;
			}
			currentSpace = (direction == BR || direction == TR) ?
					HIGH : LOW;
			break;
		default:
		}
		findMovableTile(spacePosition[highest], spaceRow[highest],
			spaceType, side);
		return true;
	}

	boolean checkingMoveNoCornTilesDir(int direction) {
		switch (direction) {
		case TR:
			return (spaceRow[ROW] != 2 * sizeY - 2 &&
			    spaceRow[TLBR] != sizeX + sizeY - 2);
		case RIGHT:
			return ((sizeY == 1 || spaceRow[TRBL] != 0) &&
			    spaceRow[TLBR] != sizeX + sizeY - 2);
		case BR:
			return (spaceRow[ROW] != 0 &&
			    spaceRow[TRBL] != 0);
		case BL:
			return (spaceRow[ROW] != 0 &&
			    spaceRow[TLBR] != 0);
		case LEFT:
			return (spaceRow[TLBR] != 0 && (sizeY == 1 ||
			    spaceRow[TRBL] != sizeX + sizeY - 2));
		case TL:
			return (spaceRow[ROW] != 2 * sizeY - 2 &&
			    spaceRow[TRBL] != sizeX + sizeY - 2);
		default:
			return false;
		}
	}

	int nextRow(int direction) {
		switch (direction) {
		case TR:
		case TL:
			return spaceRow[ROW] + 1;
		case RIGHT:
		case LEFT:
			return spaceRow[ROW];
		case BR:
		case BL:
			return spaceRow[ROW] - 1;
		default:
			return -1;
		}
	}

	boolean checkMoveNoCornTilesDir(int direction) {
		if (checkingMoveNoCornTilesDir(direction)) {
			int rowType = toRowType(direction);
			int heading = toDirHeading(direction);
			currentPosition = tileNextToSpace(rowType, heading);
			currentPositionRow = nextRow(direction);
			return true;
		}
		return false;
	}

	boolean checkMoveTilesDir(int direction) {
		if (corners)
			return checkMoveCornerTilesDir(direction);
		return checkMoveNoCornTilesDir(direction);
	}

	boolean movePuzzleDir(int direction) {
		if (checkMoveTilesDir(direction)) {
			currentDirection = direction;
			movePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	void movePuzzle(int direction, boolean control) {
		if (mouseDown || randomPaints != 0)
			return;
		if (stackPaint) {
			repaint();
			return;
		}
		if (control) {
			switch (direction) {
			case TL:
			case TOP:
			case TR:
				((HexagonsFrame) frame).callback(
					HexagonsInterface.ACTION_INCY);
				break;
			case RIGHT:
				((HexagonsFrame) frame).callback(
					HexagonsInterface.ACTION_INCX);
				break;
			case BL:
			case BOTTOM:
			case BR:
				if (sizeY <= HexagonsInterface.MIN_TILES +
						(corners ? 1 : 0))
					return;
				((HexagonsFrame) frame).callback(
					HexagonsInterface.ACTION_DECY);
				break;
			case LEFT:
				if (sizeX <= HexagonsInterface.MIN_TILES)
					return;
				((HexagonsFrame) frame).callback(
					HexagonsInterface.ACTION_DECX);
				break;
			default:
				System.out.println(
					"movePuzzle: should not get here");
			}
			return;
		}
		if (checkSolved()) {
			moveNoTiles();
			return;
		}
		if (direction == TOP || direction == BOTTOM) {
			((HexagonsFrame) frame).callback(
				HexagonsInterface.ACTION_AMBIGUOUS);
			return;
		}
		if (!movePuzzleDir(direction)) {
			((HexagonsFrame) frame).callback(
				HexagonsInterface.ACTION_BLOCKED);
		}
	}

	boolean exchangeTiles(int pos1, int pos2) {
		int tempTile;

		if (tileOfPosition[pos1] <= 0 || tileOfPosition[pos2] <= 0)
			return false;
		tempTile = tileOfPosition[pos1];
		tileOfPosition[pos1] = tileOfPosition[pos2];
		tileOfPosition[pos2] = tempTile;
		return true;
	}

	void discreteNoCornMoves(int dir, int fast) {
		int rowType, orient;

		rowType = movableNoCornTile();
		if (rowType < 0) {
			System.out.println("discreteMoves: rowType " +
				rowType);
		}
		if (dir == HIGH) {
			orient = (rowType == ROW) ? spaceRow[ROW] :
				spaceRow[ROW] - 1;
			moveNoCornTiles(tileNFromSpace(1, rowType, HIGH),
				orient, fast);
			((HexagonsFrame) frame).callback(
				HexagonsInterface.ACTION_MOVED);
			switch (rowType) {
			case TLBR:
				undo.setMove(BR);
				redo.flushMoves();
				return;
			case TRBL:
				undo.setMove(BL);
				redo.flushMoves();
				return;
			case ROW:
				undo.setMove(RIGHT);
				redo.flushMoves();
				return;
			default:
				System.out.println(
					"discreteNoCornTiles: rowType" +
					rowType);
				return;
			}
		}
		/* currentPosition > spacePosition[HIGH] */
		orient = (rowType == ROW) ? spaceRow[ROW] :
			spaceRow[ROW] + 1;
		moveNoCornTiles(tileNFromSpace(1, rowType, LOW),
			orient, fast);
		((HexagonsFrame) frame).callback(
			HexagonsInterface.ACTION_MOVED);
		switch (rowType) {
		case TLBR:
			undo.setMove(TL);
			redo.flushMoves();
			return;
		case TRBL:
			undo.setMove(TR);
			redo.flushMoves();
			return;
		case ROW:
			undo.setMove(LEFT);
			redo.flushMoves();
			return;
		default:
			System.out.println("discreteNoCornTiles: rowType" +
				rowType);
		}
	}

	int positionToTile(int positionX, int positionY) {
		int i, j, k, modI, modJ, rowPos, row;
		int x = positionX - puzzleOffset.x;
		int y = positionY - puzzleOffset.y;

		/* First convert x and y coordinates to hexagon grid.
		 * Keep in mind that the starting hexagon x position
		 * changes with "sizeX % 2". */
		if (x < tileSize.x / 4)
			return -1;
		i = 2 * (x - tileSize.x / 4) / offset.x;
		j = 3 * (y - delta.y) /
			(3 * tileSize.y / 2 + delta.y);
		modI = 2 * (x - tileSize.x / 4) % offset.x;
		modJ = 3 * (y - delta.y) %
			(3 * tileSize.y / 2 + delta.y);
		row = j / 3; /* Approximate to a rectangle just for now */
		if (j % 3 == 0) {
			/* Then it is the hexagon near bottom or top point */
			if (((sizeY - 1 + row + i) & 1) == 1) { /* \ */
				if (modJ * offset.x < modI *
					(3 * tileSize.y / 2 + delta.y))
					row--;
			} else { /* / */
				if (modJ * offset.x < (offset.x - modI) *
					(3 * tileSize.y / 2 + delta.y))
					row--;
			}
		}
		if (i < (sizeY - 1 + row) % 2 || row < 0 ||
				row > 2 * (sizeY - 1))
			return -1;
		k = (i - ((sizeY - 1 + row) % 2)) / 2;
		/* Map the hexagon grid to hexagon position in puzzle. */
		rowPos = (row < sizeY) ?
			k - (sizeY - 1 - row) / 2 :
			k + (sizeY - 1 - row) / 2;
		j = (row < sizeY) ?
			sizeX - 1 + row : 2 * (sizeY - 1) + sizeX - 1 - row;
		if (rowPos < 0 || rowPos > j)
			return -1;
		return positionFromRow(rowPos, row);
	}

	void selectCornerTiles(int space) {
		int rowType, orient;

		rowType = movableCornerTile();
		if (rowType < 0) {
			System.out.print("selectTiles: rowType = ");
			System.out.println(rowType);
			return;
		}
		orient = (spacePosition[HIGH] < spacePosition[LOW]) ?
			 ((space == HIGH) ? LOW : HIGH) : space;
		undo.setMove(findDir(currentPosition,
			spacePosition[orient], currentRow[ROW],
			spaceRow[orient]));
		redo.flushMoves();
		moveCornerTiles(currentPosition, currentRow[ROW], orient,
			NORMAL);
		started = true;
		((HexagonsFrame) frame).callback(
			HexagonsInterface.ACTION_MOVED);
		if (checkSolved()) {
			((HexagonsFrame) frame).callback(
				HexagonsInterface.ACTION_SOLVED);
		}
	}

	void selectNoCornTiles() {
		if (currentPosition < spacePosition[HIGH]) {
			if (delay > 0) {
				animateSlide(HIGH, NORMAL, true);
				if (((HexagonsFrame) frame).getToggleSound())
					((HexagonsFrame) frame).playBumpAudio();
			} else {
				while (currentPosition < spacePosition[HIGH]) {
					discreteNoCornMoves(HIGH, NORMAL);
				}
			}
		} else { /* currentPosition > spacePosition[HIGH] */
			if (delay > 0) {
				animateSlide(LOW, NORMAL, true);
				if (((HexagonsFrame) frame).getToggleSound())
					((HexagonsFrame) frame).playBumpAudio();
			} else {
				while (currentPosition > spacePosition[HIGH]) {
					discreteNoCornMoves(LOW, NORMAL);
				}
			}
		}
		if (checkSolved()) {
			((HexagonsFrame) frame).callback(
				HexagonsInterface.ACTION_SOLVED);
		}
	}

	void randomizeTiles() {
		// First interchange tiles an even number of times
		if ((sizeY > 1 && !corners) ||
				(sizeX > 1 && sizeY > 2) ||
				(sizeX == 1 && sizeY > 3)) {
			int currentPos, randomPos;
			int count = 0;

			for (currentPos = 0; currentPos < sizeSize; currentPos++) {
				randomPos = currentPos;
				if (corners && sizeX == 1 &&
				  (tileOfPosition[randomPos] == 1 ||
				  tileOfPosition[randomPos] == sizeSize - 2))
					continue;
				while (currentPos == randomPos ||
				  (corners && sizeX == 1 &&
				  (tileOfPosition[randomPos] == 1 ||
				  tileOfPosition[randomPos] == sizeSize - 2)))
					randomPos = generator.nextInt(sizeSize);
				if (exchangeTiles(currentPos, randomPos))
					count++;
			}
			if ((count & 1) == 1 && corners) {
				if (sizeX == 1) {
					if (!exchangeTiles(3, 4) &&
					    !exchangeTiles(sizeSize - 4,
							sizeSize - 5))
						System.out.println("randomizeTiles: should not get here");

				} else if (!exchangeTiles(0, 1) &&
					   !exchangeTiles(sizeSize - 2,
							sizeSize - 1))
						System.out.println("randomizeTiles: should not get here");
			}
			allTilesPaint = true;
			repaint();
		}
		// Now move the spaces around randomly
		if (sizeY > 1 || (!corners && sizeX > 1)) {
			((HexagonsFrame) frame).callback(
				HexagonsInterface.ACTION_RESET);
			lastDirection = -1;
			randomPaints = java.lang.Math.min(sizeSize,
				MAX_ITERATIONS) + generator.nextInt(2);
			repaint();
		} else if (checkSolved())
			((HexagonsFrame) frame).callback(
				HexagonsInterface.ACTION_SOLVED);
	}

	void randomizingTiles() {
		currentDirection = generator.nextInt(COORD);
		if (((currentDirection + COORD / 2) % COORD != lastDirection ||
				sizeY == 1) && checkMoveTilesDir(currentDirection)) {
			if (corners) {
				moveCornerTiles(currentPosition,
					currentPositionRow,
					((spacePosition[HIGH] < spacePosition[LOW]) ?
					((currentSpace == HIGH) ? LOW : HIGH) :
					currentSpace), INSTANT);
			} else {
				moveNoCornTiles(currentPosition,
					currentPositionRow, INSTANT);
			}
			lastDirection = currentDirection;
			currentDirection = -1;
			randomPaints--;
			try {
				Thread.sleep(delay);
			} catch (InterruptedException e) {
				// ignore
			}
		}
		if (randomPaints > 0) {
			repaint();
		} else {
			undo.flushMoves();
			redo.flushMoves();
			getStartPosition();
			((HexagonsFrame) frame).callback(
				HexagonsInterface.ACTION_RANDOMIZE);
			started = false;
			currentPosition = -1;
			if (checkSolved())
				((HexagonsFrame) frame).callback(
					HexagonsInterface.ACTION_SOLVED);
		}
	}

	void resizeTiles() {
		int i;

		for (i = 0; i <= 6; i++) {
			hexagonListX[HexagonsInterface.NOCORN][i] = tileSize.x *
				hexagonUnit[HexagonsInterface.NOCORN][i].x / 4 +
				((i > 0) ? hexagonListX[HexagonsInterface.NOCORN][i - 1] : 0);
			hexagonListY[HexagonsInterface.NOCORN][i] = 3 * tileSize.y *
				hexagonUnit[HexagonsInterface.NOCORN][i].y / 4 +
				((i > 0) ? hexagonListY[HexagonsInterface.NOCORN][i - 1] : 0);
			hexagonListX[HexagonsInterface.CORNERS][i] = tileSize.x *
				hexagonUnit[HexagonsInterface.CORNERS][i].x / 2 +
				((i > 0) ? hexagonListX[HexagonsInterface.CORNERS][i - 1] : 0);
			hexagonListY[HexagonsInterface.CORNERS][i] = tileSize.y *
				hexagonUnit[HexagonsInterface.CORNERS][i].y / 2 +
				((i > 0) ? hexagonListY[HexagonsInterface.CORNERS][i - 1] : 0);
		}
	}

	public int getDelay() {
		return delay;
	}

	public void setDelay(int delay) {
		this.delay = delay;
		/* numSlices = ((delay < MAX_SLICES) ? delay + 1 :
			MAX_SLICES); */
	}

	public boolean getCheat() {
		return cheat;
	}

	public boolean setPuzzleSizeX(int newValue) {
		if (sizeX != newValue && newValue >=
				HexagonsInterface.MIN_TILES) {
			sizeX = newValue;
			resetTiles();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	public boolean setPuzzleSizeY(int newValue) {
		if (sizeY != newValue && newValue >= ((corners) ? 1 : 0) +
				HexagonsInterface.MIN_TILES) {
			sizeY = newValue;
			resetTiles();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzleMode(boolean newValue) {
		if (corners != newValue || (!corners &&
				sizeY == HexagonsInterface.MIN_TILES)) {
			corners = newValue;
			resetTiles();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	public void speedUpPuzzle() {
		delay -= 10;
		if (delay < 0)
			delay = 0;
		((HexagonsFrame) frame).callback(
			HexagonsInterface.ACTION_SPEED);
	}

	public void slowDownPuzzle() {
		delay += 10;
		((HexagonsFrame) frame).callback(
			HexagonsInterface.ACTION_SPEED);
	}

	void showMessage(String msg) {
		((HexagonsFrame) frame).showMessage(msg);
	}

	void resizePuzzle() {
		double sqrt_3 = 1.73205080756887729352744634150587237;
		Point tempSize;

		/* Determine size of client area */
		coreSize = new Point(getSize().width, getSize().height);
		delta = new Point(5, 3);

		tileSize = new Point(java.lang.Math.max((2 * (coreSize.x +
			2 * delta.x - 1) - 2 * (sizeX + sizeY) *
			delta.x) / (2 * (sizeX + sizeY) - 1), 0),
			java.lang.Math.max((coreSize.y - 1 -
			2 * sizeY * delta.y) /
			(3 * sizeY - 1), 0));
		offset = new Point(tileSize.x + delta.x, tileSize.y + 2 * delta.y);
		tempSize = new Point((int) (offset.y * sqrt_3),
			(int) (offset.x / sqrt_3));
		if (tempSize.y < offset.y) {
			offset.x = tileSize.x + delta.x;
			offset.y = tempSize.y;
		} else {	/* tempSize.x <= offset.x */
			offset.x = tempSize.x;
			offset.y = tileSize.y + 2 * delta.y;
		}
		tileSize.x = java.lang.Math.max(offset.x - delta.x + 1, 0);
		tileSize.y = java.lang.Math.max(offset.y - 2 * delta.y - 1, 0);
		puzzleSize = new Point((sizeX + sizeY) * offset.x - tileSize.x / 2 - 2 * delta.x + 1,
			sizeY * (3 * tileSize.y + 2 * delta.y) - tileSize.y + 2);
		puzzleOffset = new Point((coreSize.x - puzzleSize.x + 2) / 2,
			(coreSize.y - puzzleSize.y + 4) / 2);
		resizeTiles();
	}

	public void initializePuzzle() {
		checkTiles();
		undo = new HexagonsStack();
		redo = new HexagonsStack();
		resetTiles();
		generator = new Random(System.nanoTime());
		hexagonUnit[HexagonsInterface.NOCORN][0] = new Point(0, 0);
		hexagonUnit[HexagonsInterface.NOCORN][1] = new Point(2, 0);
		hexagonUnit[HexagonsInterface.NOCORN][2] = new Point(1, 1);
		hexagonUnit[HexagonsInterface.NOCORN][3] = new Point(-1, 1);
		hexagonUnit[HexagonsInterface.NOCORN][4] = new Point(-2, 0);
		hexagonUnit[HexagonsInterface.NOCORN][5] = new Point(-1, -1);
		hexagonUnit[HexagonsInterface.NOCORN][6] = new Point(1, 1);
		hexagonUnit[HexagonsInterface.CORNERS][0] = new Point(0, 0);
		hexagonUnit[HexagonsInterface.CORNERS][1] = new Point(1, 1);
		hexagonUnit[HexagonsInterface.CORNERS][2] = new Point(0, 2);
		hexagonUnit[HexagonsInterface.CORNERS][3] = new Point(-1, 1);
		hexagonUnit[HexagonsInterface.CORNERS][4] = new Point(-1, -1);
		hexagonUnit[HexagonsInterface.CORNERS][5] = new Point(0, -2);
		hexagonUnit[HexagonsInterface.CORNERS][6] = new Point(1, -1);
	}

	void exposePuzzle() {
		drawFrame(false);
		drawFrame(true);
		drawAllTiles();
	}

	void selectPuzzle(int x, int y) {
		currentPosition = positionToTile(x, y);
		if (currentPosition >= 0) {
			if (checkSolved()) {
				moveNoTiles();
				currentPosition = -1;
				return;
			}
			int row = toRow(currentPosition);
			currentRow[ROW] = row;
			currentRow[TRBL] = toTrBl(currentPosition, row);
			currentRow[TLBR] = toTlBr(currentPosition, row);
			selectPaint = true;
			repaint();
		} else
			currentPosition = -1;
	}

	void releasePuzzle(int x, int y) {
		if (currentPosition == -1)
			return;
		if (corners) {
			int pos = positionToTile(x, y);
			if (pos >= 0) {
				if (spacePosition[HIGH] == pos)
					currentSpace = (spacePosition[HIGH] > spacePosition[LOW]) ?
						HIGH : LOW;
				else if (spacePosition[LOW] == pos)
					currentSpace = (spacePosition[HIGH] < spacePosition[LOW]) ?
						HIGH : LOW;
				else
					currentSpace = -1;
			} else
				currentSpace = -1;

		} else
			currentSpace = 0;
		releasePaint = true;
		repaint();
	}

	public void clearPuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		resetTiles();
		drawAllTiles();
		allTilesPaint = true;
		repaint();
		((HexagonsFrame) frame).callback(
			HexagonsInterface.ACTION_RESET);
	}

	void clearWithQueryPuzzle() {
		if (!started)
			clearPuzzle();
	}

	public void randomizePuzzle() {
		if (!mouseDown)
			randomizeTiles();
	}

	void cornersPuzzle() {
		if (mouseDown || randomPaints != 0 ||
				(sizeY == HexagonsInterface.MIN_TILES &&
				!corners))
			return;
		((HexagonsFrame) frame).callback(
			HexagonsInterface.ACTION_CORNERS);
	}

	public void getPuzzle() {
		this.stringSave = ((HexagonsFrame) frame).readTextArea();
		if (this.stringSave == null || this.stringSave.equals("")) {
			showMessage("Empty string to read");
			return;
		}
		getPuzzleString("buffer");
	}

	public void getPuzzleFile() {
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();
		try (BufferedReader br = new BufferedReader(
				new FileReader(DATAFILE))) {
			StringBuilder puzzleData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null)
				puzzleData.append(lineData).append(NL);
			this.stringSave = puzzleData.toString();
			if (this.stringSave == null || this.stringSave.equals("")) {
				showMessage("Empty file " + DATAFILE);
				return;
			}
			getPuzzleString(DATAFILE);
		} catch (IOException e) {
			//e.printStackTrace();
			showMessage("Cannot read from " + DATAFILE);
		}
	}

	public void getPuzzleString(String source) {
		try {
			int i, puzzleXDim, puzzleYDim, moves;
			boolean cornersValue;
			String symbol = SYMBOL;
			String newline = NL;
			int line = 0;

			token = this.stringSave.split(newline);
			String lineData = token[line];
			if (lineData == null || lineData.length() == 0) {
				showMessage(source +
					" nothing in buffer");
				return;
			}
			cornersValue = (Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2)) != 0);
			if (this.corners != cornersValue)
				((HexagonsFrame) frame).callback(
					HexagonsInterface.ACTION_CORNERS);
			lineData = token[++line];
			puzzleXDim = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			if (puzzleXDim >= HexagonsInterface.MIN_TILES) {
				for (i = this.sizeX; i < puzzleXDim; i++) {
					((HexagonsFrame) frame).callback(
						HexagonsInterface.ACTION_INCX);
				}
				for (i = this.sizeX; i > puzzleXDim; i--) {
					((HexagonsFrame) frame).callback(
						HexagonsInterface.ACTION_DECX);
				}
			} else {
				showMessage(source +
					" corrupted: sizeX " + puzzleXDim +
					" should be at least " +
					(HexagonsInterface.MIN_TILES +
					((cornersValue) ? 1 : 0)));
				return;
			}
			lineData = token[++line];
			puzzleYDim = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			if (puzzleYDim >= HexagonsInterface.MIN_TILES +
					((cornersValue) ? 1 : 0)) {
				for (i = this.sizeY; i < puzzleYDim; i++) {
					((HexagonsFrame) frame).callback(
						HexagonsInterface.ACTION_INCY);
				}
				for (i = this.sizeY; i > puzzleYDim; i--) {
					((HexagonsFrame) frame).callback(
						HexagonsInterface.ACTION_DECY);
				}
			} else {
				showMessage(source +
					" corrupted: sizeY " + puzzleYDim +
					" should be at least " +
					HexagonsInterface.MIN_TILES);
				return;
			}
			lineData = token[++line];
			moves = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			line = scanStartPosition(++line);
			if (line < 0)
				return;
			((HexagonsFrame) frame).callback(
				HexagonsInterface.ACTION_RESTORE);
			setStartPosition();
			allTilesPaint = true;
			repaint();
			if (!scanMoves(++line, moves))
				return;
			System.out.println(source + ": cornersValue " +
				((cornersValue) ? 1: 0) + ", sizeX " + puzzleXDim +
				", sizeY " + puzzleYDim + ", moves " + moves);
			cheat = true;
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input found when reading");
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Bad input file");
		}
	}

	public void printPuzzle(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;

		sb.append("corners").append(symbol).append(space)
			.append((corners) ? HexagonsInterface.CORNERS : HexagonsInterface.NOCORN).append(newline);
		sb.append("sizeX").append(symbol).append(space)
			.append(sizeX).append(newline);
		sb.append("sizeY").append(symbol).append(space)
			.append(sizeY).append(newline);
		sb.append("moves").append(symbol).append(space)
			.append(undo.numMoves()).append(newline);
		printStartPosition(sb);
		printMoves(sb);
	}

	public void writePuzzle() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		((HexagonsFrame) frame).writeTextArea(sb.toString());
	}

	public void writePuzzleFile() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				DATAFILE, false))) {
			pr.print(sb);
			System.out.println("Saved to " + DATAFILE);
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Cannot write to " + DATAFILE);
		}
	}

	void undoPuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0 ||
				!undo.madeMoves())
			return;
		undo.getMove();
		int direction = undo.readMoveDirection();

		redo.setMove(direction);
		// move in opposite direction
		direction = (direction + COORD / 2) % COORD;
		if (checkMoveTilesDir(direction)) {
			currentDirection = direction;
			stackPaint = true;
			repaint();
			((HexagonsFrame) frame).callback(
				HexagonsInterface.ACTION_UNDO);
		}
	}

	void redoPuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0 ||
				!redo.madeMoves())
			return;
		redo.getMove();
		int direction = redo.readMoveDirection();

		undo.setMove(direction);
		if (checkMoveTilesDir(direction)) {
			currentDirection = direction;
			stackPaint = true;
			repaint();
			((HexagonsFrame) frame).callback(
				HexagonsInterface.ACTION_REDO);
		}
	}

	void solvePuzzle() {
		showMessage("Auto-solver: sorry, not implemented.");
	}

	void printPositions() {
		int r = 0, rowPos = 0, length = 0;

		for (int i = 0; i < sizeSize; i++) {
			if (rowPos == 0) {
				if (i < sizeCenter) {
					length = sizeX + r;
					for (int tab = 0; tab < sizeX - r - 1; tab++)
						System.out.print("  ");
				} else {
					length = 3 * sizeX - r - 2;
					for (int tab = 0; tab < r - sizeX + 1; tab++)
						System.out.print("  ");
				}
			}
			int pos = startPosition[i];
			if (pos < 100)
				System.out.print(" ");
			if (pos < 10 && pos >= 0)
				System.out.print(" ");
			System.out.print(pos);
			if (rowPos == length - 1) {
				r++;
				rowPos = 0;
				System.out.println();
			} else {
				rowPos++;
				System.out.print(" ");
			}
		}
		System.out.println();
	}

	void getStartPosition() {
		startSpacePosition[LOW] = spacePosition[LOW];
		startSpacePosition[HIGH] = spacePosition[HIGH];
		startSpaceRow[TRBL] = spaceRow[TRBL];
		startSpaceRow[TLBR] = spaceRow[TLBR];
		startSpaceRow[ROW] = spaceRow[ROW];
		System.arraycopy(tileOfPosition, 0, startPosition, 0, sizeSize);
	}

	void setStartPosition() {
		spacePosition[LOW] = startSpacePosition[LOW];
		spacePosition[HIGH] = startSpacePosition[HIGH];
		spaceRow[TRBL] = startSpaceRow[TRBL];
		spaceRow[TLBR] = startSpaceRow[TLBR];
		spaceRow[ROW] = startSpaceRow[ROW];
		System.arraycopy(startPosition, 0, tileOfPosition, 0, sizeSize);
		// drawAll
	}

	int scanStartPosition(int location) {
		String lineData;
		int line = location;

		try {
			String space = "\\s+";
			int n = 0, r = 0, length;

			++line; //read startingPosition:
			line++;
			if (line >= token.length) {
				System.out.println("Short data in StartPosition");
				return -1;
			}
			lineData = token[line];
			String[] st = lineData.split(space);
			int offset = ("".equals(st[0])) ? 1 : 0;
			for (int i = 0; i < sizeSize; i++) {
				if (i < sizeCenter) {
					length = sizeX + r;
				} else {
					length = 2 * sizeY + sizeX - r - 2;
				}
				int num = Integer.parseInt(st[n + offset].trim());
				startPosition[i] = num;
				if (num == 0)
					startSpacePosition[HIGH] = i;
				else if (num == -1)
					startSpacePosition[LOW] = i;
				n++;
				if (n == length &&
						i != sizeSize - 1) {
					r++;
					n = 0;
					line++;
					if (line >= token.length) {
						System.out.println("Short data in StartPosition");
						return -1;
					}
					lineData = token[line];
					st = lineData.split(space);
				}
			}
			//blank line read
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input in StartPosition");
			return -1;
		}
		if (corners) {
			startSpaceRow[HIGH] = toRow(startSpacePosition[HIGH]);
			if (sizeX > 1 || sizeY > 1)
				startSpaceRow[LOW] = toRow(startSpacePosition[LOW]);
		} else {
			startSpaceRow[ROW] = toRow(startSpacePosition[HIGH]);
			startSpaceRow[TRBL] = toTrBl(startSpacePosition[HIGH],
				startSpaceRow[ROW]);
			startSpaceRow[TLBR] = toTlBr(startSpacePosition[HIGH],
				startSpaceRow[ROW]);
		}
		return line;
	}

	boolean scanMoves(int location, int moves) {
		Graphics g = getGraphics();

		if (g == null)
			return false;
		try {
			int l;
			int line = location;
			String lineData;
			String symbol = SYMBOL;

			++line; //read moves direction
			for (l = 0; l < moves && ++line < token.length; l++) {
				lineData = token[line];
				currentDirection = Integer.parseInt(
					lineData.substring(lineData.indexOf(symbol) +
					2));
				if (checkMoveTilesDir(currentDirection)) {
					scanPaint = true;
					paint(g);
				} else {
					System.out.println(l +
						": move in direction " +
						currentDirection +
						", cannot be made.");
					currentDirection = -1;
					return false;
				}

			}
			if (l < moves) {
				showMessage("Corrupt input in scanMoves, only " + l + " moves");
				return false;
			}
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input found when reading (scanMoves)");
			return false;
		} finally {
			g.dispose();
		}
		return true;
	}

	void printStartPosition(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;
		int r = 0, rowPos = 0, length = 0;

		sb.append(newline);
		sb.append("startingPosition").append(symbol).append(newline);
		for (int i = 0; i < sizeSize; i++) {
			if (rowPos == 0) {
				if (i < sizeCenter) {
					length = sizeX + r;
					for (int tab = 0; tab < sizeY - r - 1; tab++)
						sb.append(space).append(space);
				} else {
					length = 2 * sizeY + sizeX - r - 2;
					for (int tab = 0; tab < r - sizeY + 1; tab++)
						sb.append(space).append(space);
				}
			}
			int pos = startPosition[i];
			if (pos < 100)
				sb.append(space);
			if (pos < 10 && pos >= 0)
				sb.append(space);
			sb.append(space).append(pos);
			if (rowPos == length - 1) {
				r++;
				rowPos = 0;
				sb.append(newline);
			} else {
				rowPos++;
			}
		}
		sb.append(newline);
	}

	void printMoves(StringBuilder sb) {
		int direction;
		int nMoves = undo.numMoves();
		String symbol = SYMBOL;
		String space = "\t";
		String newline = NL;

		sb.append("moves").append(space)
			.append("direction").append(newline);
		while (undo.madeMoves()) {
			undo.getMove();
			direction = undo.readMoveDirection();
			redo.setMove(direction);
		}
		// apply all moves back to the stack since
		// we emptied it above!!
		for (int i = 0; i < nMoves; i++) {
			redo.getMove();
			direction = redo.readMoveDirection();
			sb.append(i + 1).append(symbol).append(space)
				.append(direction).append(newline);
			undo.setMove(direction);
		}
	}

	@Override
	public void paint(Graphics g) {
		if (firstPaint) {
			initializePuzzle();
			firstPaint = false;
			resizePaint = true;
		}
		if (!framePaint && !allTilesPaint &&
				!selectPaint && !releasePaint && !movePaint &&
				!scanPaint && !stackPaint && randomPaints == 0) 
			resizePaint = true;
		if (resizePaint) {
			resizePuzzle();
			exposePuzzle();
			resizePaint = false;
		}
		if (framePaint) {
			drawFrame(true);
			framePaint = false;
		}
		if (allTilesPaint) {
			drawAllTiles();
			allTilesPaint = false;
		}
		if (selectPaint) {
			if (currentPosition != -1) {
				int rowType = (corners) ? movableCornerTile() :
				  movableNoCornTile();

				if (rowType < 0) {
				  drawTile(currentPosition,
				    rowType == SPACE ,
				    false, 1, 0, 0);
				  try {
				    Thread.sleep((long) 3 * delay);
				  } catch (InterruptedException e) {
					  // ignore
				  }
				  drawTile(currentPosition,
				    true, true, 1, 0, 0);
				  if (rowType != SPACE)
				    drawTile(currentPosition,
				      false, false, 0, 0, 0);
				  ((HexagonsFrame) frame).callback(rowType);
				  currentPosition = -1;
				  mouseDown = false;
				} else
				  drawTile(currentPosition,
				    false, false, 1, 0, 0);
			}
			selectPaint = false;
		}
		if (releasePaint) {
			if (currentPosition != -1) {
				drawTile(currentPosition,
					true, true, 1, 0, 0);
				drawTile(currentPosition,
					false, false, 0, 0, 0);
				if (currentSpace != -1)
					if (corners)
						selectCornerTiles(currentSpace);
					else
						selectNoCornTiles();
				currentPosition = -1;
			}
			releasePaint = false;
		}
		if (movePaint || scanPaint || stackPaint) {
			int fast;

			if (stackPaint)
				fast = DOUBLE;
			else if (movePaint)
				fast = NORMAL;
			else
				fast = INSTANT;
			if (corners)
				moveCornerTiles(currentPosition,
					currentPositionRow,
					((spacePosition[HIGH] < spacePosition[LOW]) ?
					((currentSpace == HIGH) ? LOW : HIGH) :
					currentSpace), fast);
			else
				moveNoCornTiles(currentPosition,
					currentPositionRow, fast);
			started = true;
			if (!stackPaint && (movePaint || scanPaint)) {
				((HexagonsFrame) frame).callback(
					HexagonsInterface.ACTION_MOVED);
				undo.setMove(currentDirection);
				redo.flushMoves();
				if (checkSolved())
					((HexagonsFrame) frame).callback(
						HexagonsInterface.ACTION_SOLVED);
				movePaint = false;
				scanPaint = false;
			}
			stackPaint = false;
			currentDirection = -1;
			currentPosition = -1;
		}
		if (randomPaints > 0) {
			randomizingTiles();
		}
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}
}
